#ifndef DUNE_TECTONIC_LOCAL_NONLINEARITY_HH
#define DUNE_TECTONIC_LOCAL_NONLINEARITY_HH

#include <cmath>
#include <limits>

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>
#include <dune/common/shared_ptr.hh>

#include <dune/fufem/interval.hh>

#include "nicefunction.hh"

namespace Dune {
template <int dimension> class LocalNonlinearity {
public:
  typedef FieldVector<double, dimension> VectorType;
  typedef FieldMatrix<double, dimension, dimension> MatrixType;

  LocalNonlinearity(shared_ptr<NiceFunction const> func) : func(func) {}

  double operator()(VectorType const &x) const {
    double ret;
    func->evaluate(x.two_norm(), ret);
    return ret;
  }

  double regularity(VectorType const &x) const {
    if (!func->smoothesNorm() && x.two_norm() < 1e-14) // TODO: Make this
                                                       // controllable
                                                       // (truncationRadius?)
      return std::numeric_limits<double>::infinity();

    return func->regularity(x.two_norm());
  }

  // directional subdifferential: at u on the line u + t*v
  // u and v are assumed to be non-zero
  void directionalSubDiff(VectorType const &u, VectorType const &v,
                          Interval<double> &D) const {
    if (u.two_norm() == 0) {
      D[0] = D[1] = func->rightDifferential(0) * v.two_norm();
      return;
    }
    double const un = u.two_norm();
    double const ndotp = (u * v) / un;
    if (ndotp > 0) {
      D[1] = ndotp * func->rightDifferential(un);
      D[0] = ndotp * func->leftDifferential(un);
    } else {
      D[1] = ndotp * func->leftDifferential(un);
      D[0] = ndotp * func->rightDifferential(un);
    }
  }

  /** Formula for the derivative:

    \f{align*}{
      \frac {d^2}{dz^2} H(|z|)
      &= \frac d{dz} \left( H'(|z|) \otimes \frac z{|z|} \right)\\
      &= H''(|z|) \frac z{|z|} \otimes \frac z{|z|}
      + H'(|z|) \left( \frac {|z| \operatorname{id} - z \otimes z/|z|}{|z|^2}
    \right)\\
      &= \frac {H''(|z|)}{|z|^2} z \otimes z
      + \frac {H'(|z|)}{|z|} \operatorname{id}
      - \frac {H'(|z|)}{|z|^3} z \otimes z\\
      &= \left( \frac {H''(|z|)}{|z|^2} - \frac {H'(|z|)}{|z|^3} \right) z
    \otimes z
      + \frac {H'(|z|)}{|z|} \operatorname{id}
    \f}
  */
  void addHessian(VectorType const &x, MatrixType &A) const {
    double const normX2 = x.two_norm2();
    double const normX = std::sqrt(normX2);
    double const normX3 = normX * normX2;

    double const H1 = func->rightDifferential(normX);
    double const H2 = func->second_deriv(normX);

    // TODO: potential optimisation: factor out (H1 / normX), get rid of normX3
    double const weight1 = H2 / normX2;
    double const weight2 = -H1 / normX3;
    double const weight3 = H1 / normX;

    // {{{ In what follows, we handle the case 0 * (1/x) = 0 with x
    // close to 0.
    double tensorweight = 0;

    if (std::isnan(weight1))
      assert(H2 == 0);
    else
      tensorweight += weight1;

    if (std::isnan(weight2))
      assert(H1 == 0);
    else
      tensorweight += weight2;

    double idweight = 0;
    if (std::isnan(weight3))
      assert(H1 == 0);
    else
      idweight += weight3;
    // }}}

    for (int i = 0; i < dimension; ++i)
      for (int j = 0; j < i; ++j) {
        double const entry = tensorweight * x[i] * x[j];
        A[i][j] += entry;
        A[j][i] += entry;
      }

    for (int k = 0; k < dimension; ++k) {
      double const entry = tensorweight * x[k] * x[k];
      A[k][k] += entry + idweight;
    }
  }

  void addGradient(VectorType const &x, VectorType &y) const {
    double const xnorm = x.two_norm();
    if (xnorm == 0 or std::isinf(func->regularity(xnorm)))
      return;

    // left and right differential coincide in this case
    y.axpy(func->rightDifferential(xnorm) / xnorm, x);
  }

  void upperGradient(VectorType const &x, VectorType &ret) const {
    double const xnorm = x.two_norm();
    assert(xnorm != 0);

    ret = x;
    ret *= func->rightDifferential(xnorm) / xnorm;
  }

  void lowerGradient(VectorType const &x, VectorType &ret) const {
    double const xnorm = x.two_norm();
    assert(xnorm != 0);

    ret = x;
    ret *= func->leftDifferential(xnorm) / xnorm;
  }

  void directionalDomain(VectorType const &, VectorType const &,
                         Interval<double> &dom) const {
    dom[0] = -std::numeric_limits<double>::max();
    dom[1] = std::numeric_limits<double>::max();
  }

private:
  shared_ptr<NiceFunction const> const func;
};
}
#endif
