#ifndef NICE_FUNCTION_HH
#define NICE_FUNCTION_HH

#include <algorithm>
#include <cmath>
#include <limits>

#include <dune/common/exceptions.hh>
#include <dune/common/function.hh>

namespace Dune {
class NiceFunction {
public:
  virtual ~NiceFunction() {}

  double virtual leftDifferential(double s) const = 0;
  double virtual rightDifferential(double s) const = 0;

  double virtual second_deriv(double x) const = 0;

  double virtual regularity(double s) const = 0;

  // Whether H(|.|) is smooth at zero
  bool virtual smoothesNorm() const { return false; }

  void virtual evaluate(double const &x, double &y) const {
    DUNE_THROW(NotImplemented, "evaluation not implemented");
  }
};

class RuinaFunction : public NiceFunction {
public:
  RuinaFunction(double coefficient, double a, double mu, double eta,
                double normalStress, double b, double state, double L, double h)
      : a(a),
        eta(eta),
        h(h),
        coefficientProduct(coefficient * normalStress),
        K(mu + b * (state - std::log(eta * L))), // state is assumed to be
                                                 // logarithmic
        rho(std::exp(-K / a)) {}

  double virtual leftDifferential(double x) const {
    double const arg = x / h * eta;
    if (arg <= rho)
      return 0;

    double const ret = a * std::log(arg) + K;
    return coefficientProduct * ret;
  }

  double virtual rightDifferential(double s) const {
    return leftDifferential(s);
  }

  double virtual second_deriv(double x) const {
    double const arg = x / h * eta;
    if (arg <= rho)
      return 0;

    return coefficientProduct * a / x;
  }

  double virtual regularity(double x) const {
    double const arg = x / h * eta;
    // TODO: Make this controllable
    if (std::abs(arg - rho) < 1e-14)
      return std::numeric_limits<double>::infinity();

    return std::abs(second_deriv(x));
  }

  bool virtual smoothesNorm() const { return true; }

private:
  double const a;
  double const eta;
  double const h;
  double const coefficientProduct;
  double const K;
  double const rho;
};

class TrivialFunction : public NiceFunction {
public:
  void virtual evaluate(double const &x, double &y) const { y = 0; }

  double virtual leftDifferential(double) const { return 0; }

  double virtual rightDifferential(double) const { return 0; }

  double virtual second_deriv(double) const { return 0; }

  double virtual regularity(double) const { return 0; }

  bool virtual smoothesNorm() const { return true; }
};
}
#endif
